package ch.inf3.indexation.crawler;

import ch.inf3.indexation.model.Bookmark;
import java.io.IOException;
import java.net.NoRouteToHostException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import javax.net.ssl.SSLException;
import javax.persistence.EntityManager;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.client.solrj.impl.HttpSolrClient;
import org.apache.solr.client.solrj.response.UpdateResponse;
import org.apache.solr.common.SolrInputDocument;
import org.jsoup.Connection;
import org.jsoup.Connection.Response;
import org.jsoup.HttpStatusException;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;

public class PageCrawler implements Callable<Boolean> {

    private static final Logger LOG = Logger.getLogger(PageCrawler.class.getName());

    private final EntityManager em;

    private final String bookmarkId;
    private final HttpSolrClient solr;
    private final ConcurrentHashMap<String, ?> currentJobs;

    public PageCrawler(EntityManager em, String bookmarkId, HttpSolrClient solr, ConcurrentHashMap<String, ?> currentJobs) {
        this.em = em;
        this.bookmarkId = bookmarkId;
        this.solr = solr;
        this.currentJobs = currentJobs;

        LOG.log(Level.INFO, "Building Crawler for bookmark {0}", bookmarkId);
    }

    @Override
    public Boolean call() throws Exception {
        Bookmark bookmarkEntity = null;
        String uri = null;
        try {
            em.getTransaction().begin();

            bookmarkEntity = em.find(Bookmark.class, bookmarkId);
            if (bookmarkEntity == null) {
                LOG.log(Level.WARNING, "Bookmark {0} not found during crawling.", bookmarkId);
                currentJobs.remove(bookmarkId);
                return false;
            }

            uri = bookmarkEntity.getUri();

            LOG.log(Level.INFO, "Got URI {0} for bookmark {1}", new Object[]{uri, bookmarkId});

            Response response = Jsoup.connect(uri)
                    .followRedirects(true)
                    .method(Connection.Method.GET)
                    .execute();

            Document doc = response.parse();
            String lastUri = doc.location();
            if (!uri.equals(lastUri)) {
                bookmarkEntity.setLastUri(lastUri);
            }

            bookmarkEntity.setLastRetrieved(new Date());
            bookmarkEntity.setLastStatus(response.statusCode());

            LOG.log(Level.FINE, "Response is at {0}", lastUri);
            LOG.log(Level.FINER, "Headers are {0}", response.headers());

            Map<String, List<String>> metaHtml = doc.head().select("meta").stream()
                    .collect(Collectors.groupingBy(
                            e -> e.attr("property").isEmpty() ? e.attr("name") : e.attr("property"),
                            Collectors.mapping(e -> e.attr("content"), Collectors.toList())
                    ));
            LOG.log(Level.FINER, "Meta are {0}", metaHtml);


            SolrInputDocument solrDoc = new SolrInputDocument();
            solrDoc.addField("id", bookmarkEntity.getUuid());
            solrDoc.addField("id_i", bookmarkEntity.getId());
            solrDoc.addField("guid_s", bookmarkEntity.getGuid());

            solrDoc.addField("title_txt", doc.title());
            solrDoc.addField("content_txt", doc.body().text());

            UpdateResponse indexResponse = solr.add(solrDoc, 5000);

            return true;
        } catch (HttpStatusException ex) {
            LOG.log(Level.WARNING, "Got a response with a non-success code: ", ex);

            if (bookmarkEntity != null) {
                bookmarkEntity.setLastRetrieved(new Date());
                bookmarkEntity.setLastStatus(ex.getStatusCode());
            }

            return false;
        } catch (SocketTimeoutException ex) {
            LOG.log(Level.WARNING, "Got a timeout during connection to {0}", uri);

            if (bookmarkEntity != null) {
                bookmarkEntity.setLastRetrieved(new Date());
                bookmarkEntity.setLastStatus(900);
            }

            return false;
        } catch (UnknownHostException | NoRouteToHostException ex) {
            LOG.log(Level.WARNING, "Unknown host at {0}", uri);

            if (bookmarkEntity != null) {
                bookmarkEntity.setLastRetrieved(new Date());
                bookmarkEntity.setLastStatus(901);
            }

            return false;
        } catch (SSLException ex) {
            LOG.log(Level.WARNING, "SSL Error while connecting to {0}", uri);

            if (bookmarkEntity != null) {
                bookmarkEntity.setLastRetrieved(new Date());
                bookmarkEntity.setLastStatus(902);
            }

            return false;
        } catch (IOException | SolrServerException ex) {
            LOG.log(Level.WARNING, "Exception during query of {0}", bookmarkId);
            LOG.log(Level.INFO, "Exception is: ", ex);
            return false;
        } finally {
            em.getTransaction().commit();

            if (currentJobs != null) {
                currentJobs.remove(bookmarkId);
            }
        }
    }

}
