package ch.inf3.indexation.protocol;

import ch.inf3.indexation.model.Bookmark;
import ch.inf3.indexation.server.BookmarkController;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.eclipse.microprofile.openapi.annotations.media.Schema;

@Schema(
        name = "SearchResponse",
        description = "POJO that represents the search results."
)
public class SearchResponse {

    @Schema(
            name = "SearchResultEntry",
            description = "POJO that represents one search result."
    )
    public static class SearchResultEntry {
        private final String uuid;
        private final String guid;
        private final String uri;
        private final String prefix;
        private final String highlightTitle;
        private final String highlightContent;
        private final Date lastRetrieved;

        public SearchResultEntry(Bookmark bookmarkEntity, String highlightTitle, String highlightContent) {
            this.uuid = bookmarkEntity.getUuid();
            this.guid = bookmarkEntity.getGuid();
            this.uri = bookmarkEntity.getLastUri() != null ? bookmarkEntity.getLastUri() : bookmarkEntity.getUri();

            this.prefix = bookmarkEntity.getInternalPrefix();
            this.lastRetrieved = bookmarkEntity.getLastRetrieved();

            this.highlightTitle = highlightTitle;
            this.highlightContent = highlightContent;
        }

        public String getUuid() {
            return uuid;
        }

        public String getGuid() {
            return guid;
        }

        public String getUri() {
            return uri;
        }

        public String getPrefix() {
            return prefix;
        }

        public String getHighlightTitle() {
            return highlightTitle;
        }

        public String getHighlightContent() {
            return highlightContent;
        }

        public Date getLastRetrieved() {
            return lastRetrieved;
        }
    }

    private final Collection<SearchResultEntry> results;
    private final Integer resultCount;
    private final Integer searchTime;

    public SearchResponse(Collection<Bookmark> bookmarkEntities, QueryResponse searchResult) {
        searchTime = searchResult.getQTime();
        resultCount = bookmarkEntities.size();

        results = new ArrayList<>(bookmarkEntities.size());

        Map<String, Map<String, List<String>>> highlights = searchResult.getHighlighting();

        for (Bookmark bookmarkEntity : bookmarkEntities) {
            Map<String, List<String>> highlightEntry = highlights.getOrDefault(bookmarkEntity.getUuid(), Collections.EMPTY_MAP);

            Collection<String> highlightsTitle = highlightEntry.getOrDefault(BookmarkController.SOLR_FIELD_TITLE, Collections.EMPTY_LIST);
            Collection<String> highlightsContent = highlightEntry.getOrDefault(BookmarkController.SOLR_FIELD_CONTENT, Collections.EMPTY_LIST);

            String highlightTitle = highlightsTitle.stream().collect(Collectors.joining(" ... "));
            String highlightContent = highlightsContent.stream().collect(Collectors.joining(" ... "));

            results.add(new SearchResultEntry(bookmarkEntity, highlightTitle, highlightContent));
        }
    }

    public Collection<SearchResultEntry> getResults() {
        return results;
    }

    public Integer getResultCount() {
        return resultCount;
    }

    public Integer getSearchTime() {
        return searchTime;
    }
}
