package ch.inf3.indexation.server;

import ch.inf3.indexation.importer.BookmarkJsonPage;
import ch.inf3.indexation.importer.BookmarksJsonImporter;
import ch.inf3.indexation.model.Bookmark;
import ch.inf3.indexation.protocol.SearchResponse;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.json.JsonException;
import javax.persistence.EntityManager;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import org.apache.commons.fileupload.FileItemIterator;
import org.apache.commons.fileupload.FileItemStream;
import org.apache.commons.fileupload.FileUploadException;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.SolrServerException;
import org.apache.solr.client.solrj.impl.HttpSolrClient;
import org.apache.solr.client.solrj.response.QueryResponse;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.eclipse.microprofile.metrics.annotation.Timed;
import org.eclipse.microprofile.openapi.annotations.Operation;
import org.eclipse.microprofile.openapi.annotations.enums.SchemaType;
import org.eclipse.microprofile.openapi.annotations.media.Content;
import org.eclipse.microprofile.openapi.annotations.media.Schema;
import org.eclipse.microprofile.openapi.annotations.parameters.Parameter;
import org.eclipse.microprofile.openapi.annotations.responses.APIResponse;
import org.eclipse.microprofile.openapi.annotations.responses.APIResponses;

@Path("/bookmark")
@RequestScoped
public class BookmarkController {

    private static final Logger LOG = Logger.getLogger(BookmarkController.class.getName());

    public static final String SOLR_FIELD_CONTENT = "content_txt";
    public static final String SOLR_FIELD_TITLE = "title_txt";
    public static final String SOLR_HIGHLIGHT_PRE = "<span>";
    public static final String SOLR_HIGHLIGHT_POST = "</span>";

    @Inject
    private EntityManager em;

    @Inject
    private HttpSolrClient solr;

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Timed(name = " bookmark-list")
    public Response listBookmarks() {
        em.getTransaction().begin();

        List<Bookmark> todos = em.createNamedQuery("Bookmark.findAll", Bookmark.class).getResultList();

        em.getTransaction().commit();

        return Response
                .ok(todos)
                .build();
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path("/{bookmarkId}")
    public Response getSingle(@PathParam("bookmarkId") String bookmarkId) {
        if (bookmarkId == null || bookmarkId.isBlank()) {
            return Response
                    .status(Status.BAD_REQUEST)
                    .build();
        }

        Bookmark found = em.find(Bookmark.class, bookmarkId);

        if (found == null) {
            return Response
                    .status(Status.NOT_FOUND)
                    .build();
        }

        return Response
                .ok(found)
                .build();
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path("/search")
    @APIResponses({
        @APIResponse(
                responseCode = "200",
                description = "A search result of the given search term.",
                content = @Content(
                        mediaType = "application/json",
                        schema = @Schema(implementation = SearchResponse.class)
                )
        )
    })
    @Operation(
            summary = "Search for bookmarks containing a specified text either in the title or the content."
    )
    public Response search(
            @Parameter(description = "Search term applied on the page title.")
            @QueryParam("title") String title,
            @Parameter(description = "Search term applied on the page content.")
            @QueryParam("text") String text,
            @Parameter(description = "Search term applied on both the page title and content.")
            @QueryParam("q") String all) {

        Collection<String> query = new ArrayList<>(4);

        if (title != null && !title.isEmpty()) {
            query.add(SOLR_FIELD_TITLE + ":\"" + title + "\"");
        }

        if (text != null && !text.isEmpty()) {
            query.add(SOLR_FIELD_CONTENT + ":\"" + text + "\"");
        }

        if (all != null && !all.isEmpty()) {
            query.add(SOLR_FIELD_TITLE + ":\"" + all + "\" OR " + SOLR_FIELD_CONTENT + ":\"" + all + "\"");
        }

        SolrQuery params = new SolrQuery(query.stream().collect(Collectors.joining(" AND ")));

        params.setHighlight(true);
        params.addHighlightField(SOLR_FIELD_TITLE);
        params.addHighlightField(SOLR_FIELD_CONTENT);
        params.setHighlightSimplePre(SOLR_HIGHLIGHT_PRE);
        params.setHighlightSimplePost(SOLR_HIGHLIGHT_POST);

        try {
            QueryResponse searchResult = solr.query(params);
            SolrDocumentList results = searchResult.getResults();

            Collection<Bookmark> bookmarkEntities = new ArrayList<>(results.size());

            for (SolrDocument result : results) {
                String documentId = (String) result.get("id");

                Bookmark found = em.find(Bookmark.class, documentId);

                if (found == null) {
                    continue;
                }

                bookmarkEntities.add(found);
            }

            return Response
                    .ok(new SearchResponse(bookmarkEntities, searchResult))
                    .build();
        } catch (SolrServerException | IOException ex) {
            LOG.log(Level.SEVERE, "Exception while querying Solr: ", ex);
            return Response
                    .status(Status.INTERNAL_SERVER_ERROR)
                    .entity(ex.getMessage())
                    .build();
        }
    }

    @DELETE
    @Produces(MediaType.TEXT_PLAIN)
    @Path("/{bookmarkId}")
    @APIResponses({
        @APIResponse(
                responseCode = "200",
                description = "Validation that the bookmark has been deleted.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "Deleted 522eac0c-a81b-4148-8443-54e3243e1bf4"
                )
        ),
        @APIResponse(
                responseCode = "404",
                description = "If the bookmark specified was not found.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "Bookmark 522eac0c-a81b-4148-8443-54e3243e1bf4 not found."
                )
        ),
        @APIResponse(
                responseCode = "400",
                description = "If the given identifier is invalid or empty.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "No bookmark identifier specified."
                )
        )
    })
    @Operation(
            summary = "Delete a bookmark by its UUID."
    )
    public Response clearSingle(
            @Parameter(
                    description = "UUID of the bookmark to delete",
                    required = true,
                    example = "522eac0c-a81b-4148-8443-54e3243e1bf4",
                    schema = @Schema(type = SchemaType.STRING)
            )
            @PathParam("bookmarkId") String bookmarkId
    ) {
        if (bookmarkId == null || bookmarkId.isBlank()) {
            return Response
                    .status(Status.BAD_REQUEST)
                    .entity("No bookmark specified")
                    .build();
        }

        LOG.log(Level.INFO, "Trying to delete {0}", bookmarkId);

        em.getTransaction().begin();
        Bookmark found = em.find(Bookmark.class, bookmarkId);

        if (found == null) {
            em.getTransaction().rollback();
            return Response
                    .status(Status.NOT_FOUND)
                    .entity("Bookmark not found")
                    .build();
        }

        em.remove(found);
        em.getTransaction().commit();

        return Response
                .ok("Deleted " + found.getGuid() + " bookmark")
                .build();
    }

    @POST
    @Path("/upload")
    @Consumes(MediaType.MULTIPART_FORM_DATA)
    @Produces(MediaType.TEXT_PLAIN)
    @Timed(name = "bookmark-parsing")
    @APIResponses({
        @APIResponse(
                responseCode = "200",
                description = "The validation of the import.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "Imported 2323 records"
                )
        ),
        @APIResponse(
                responseCode = "400",
                description = "The uploaded file is not a valid JSON document or "
                + "contains an unexpected structure.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "JSON error in the uploaded file: element cannot be "
                        + "mapped to a null key."
                )
        ),
        @APIResponse(
                responseCode = "500",
                description = "An internal error occurred during the connexion with "
                + "the database.",
                content = @Content(
                        mediaType = "text/plain",
                        schema = @Schema(type = SchemaType.STRING),
                        example = "Error while reading the uploaded file: the request "
                        + "was rejected because its size exceeds the configured maximum."
                )
        )
    })
    @Operation(
            summary = "Import the content of a Firefox JSON bookmarks export file.",
            description = "The file can be uploaded with CuRL with the following "
            + "command: curl -F \"file=@bookmarks-2019-11-01.json\" localhost:8080/api/bookmarks/upload"
    )
    public Response uploadTest(@Context HttpServletRequest request) {
        try {
            ServletFileUpload upload = new ServletFileUpload();
            FileItemIterator itemIterator = upload.getItemIterator(request);

            int totalItems = 0;
            while (itemIterator.hasNext()) {
                FileItemStream item = itemIterator.next();
                InputStream is = item.openStream();

                BookmarksJsonImporter jsonImporter = new BookmarksJsonImporter(is);

                Collection<BookmarkJsonPage> jsonEntries = jsonImporter.getEntries();
                totalItems += jsonEntries.size();

                em.getTransaction().begin();

                for (BookmarkJsonPage jsonEntry : jsonEntries) {
                    List<Bookmark> found = em.createQuery("SELECT t FROM Bookmark t WHERE t.guid = ?1")
                            .setParameter(1, jsonEntry.getGuid())
                            .getResultList();

                    Bookmark b;
                    if (found.isEmpty()) {
                        LOG.log(Level.FINER, "Creating new Bookmark {0}", jsonEntry.getGuid());
                        b = jsonEntry.toEntity();
                        em.persist(b);
                    } else {
                        LOG.log(Level.FINER, "Updating existing Bookmark {0}", jsonEntry.getGuid());
                        b = found.get(0);
                        jsonEntry.updateEntity(b);
                    }
                }

                em.getTransaction().commit();
            }

            return Response
                    .ok("Parsed " + totalItems + " records\r\n")
                    .build();
        } catch (JsonException ex) {
            return Response
                    .status(Status.BAD_REQUEST)
                    .entity("JSON error in the uploaded file: " + ex.getMessage())
                    .build();
        } catch (IOException | FileUploadException ex) {
            return Response
                    .status(Status.INTERNAL_SERVER_ERROR)
                    .entity("Error while reading the file: " + ex.getMessage())
                    .build();
        }
    }

}
